import stamp.core.*;

/*
  Class which communicates with CMUCamera to track objects
  of a particular color.
*/
class CMUCam
{
  public CMUCam(int nRxPin, int nTxPin)
  {
    m_uartRx        = new Uart(
                    Uart.dirReceive,
                    nRxPin,
                    Uart.dontInvert,
                    BAUD_RATE,
                    STOP_BITS );
    m_uartTx        = new Uart(
                      Uart.dirTransmit,
                      nTxPin,
                      Uart.dontInvert,
                      BAUD_RATE,
                      STOP_BITS );
    m_timer         = new Timer();
    m_buffTemp      = new StringBuffer(32);
    m_rgchBuff      = new char[8];
    m_fInit         = false;
    m_fRGB          = true;
    m_fAutoWhite    = false;
    m_nFrameSpeed   = FRAME_SPEED_17;
    m_nX1           = 1;
    m_nY1           = 1;
    m_nX2           = 80;
    m_nY2           = 143;
    m_nTrackingMode = TRACKING_NONE;

    FullReset();
  }

  protected char receiveByte()
  {
    int n = 0xff & m_uartRx.receiveByte();
    System.out.print(n);
    System.out.print(" ");
    System.out.println((char)n);

    return (char) n;
  }

  // Full Reset Camera into RAW mode
  public boolean FullReset()
  {
    // Send CR to make sure that camera is in idle mode
    m_uartTx.sendString("\r");
    if ( !WaitForPrompt() )
    {
      return false;
    }

    // Reset the camera
    m_uartTx.sendString("RS\r");
    if ( !WaitForPrompt() )
    {
      return false;
    }

    // Configure camera for raw mode
    m_uartTx.sendString("RM 1\r");
    if ( !WaitForACKPrompt() )
    {
      return false;
    }

    m_fInit = true;
    return true;
  }

  // Reset Camera into Idle mode
  public boolean Reset()
  {
    // Check to make sure that camera has received a full reset
    if ( !m_fInit )
    {
      if ( !FullReset() )
      {
        return false;
      }
    }

    // Send CR to make sure that camera is in idle mode
    m_uartTx.sendString("\r");
    if ( !WaitForPrompt() )
    {
      return false;
    }

    // A reset will bring us out of any tracking that might be
    // taking place
    m_nTrackingMode = TRACKING_NONE;
    return true;
  }

  // Set Specific Camera Registers such as auto gain, etc.
  public boolean SetContrast(int nContrast)
  {
    if (nContrast > 255 || nContrast < 0 )
    {
      return false;
    }

    // Make sure camera is in reset mode
    if ( !Reset() )
    {
      return false;
    }

    // Configure camera contrast
    m_buffTemp.clear();
    m_buffTemp.append("CR 5 ");
    m_buffTemp.append(nContrast);
    m_buffTemp.append("\r");

    m_uartTx.sendString(m_buffTemp.toString());
    if ( !WaitForACKPrompt() )
    {
      return false;
    }

    return true;
  }

  public boolean SetBrightness(int nBrightness)
  {
    if (nBrightness > 255 || nBrightness < 0 )
    {
      return false;
    }

    // Make sure camera is in reset mode
    if ( !Reset() )
    {
      return false;
    }

    // Configure camera brightness
    m_buffTemp.clear();
    m_buffTemp.append("CR 6 ");
    m_buffTemp.append(nBrightness);
    m_buffTemp.append("\r");

    m_uartTx.sendString(m_buffTemp.toString());
    if ( !WaitForACKPrompt() )
    {
      return false;
    }

    return true;
  }

  public boolean SetColorMode(boolean fRGB, boolean fAutoWhite)
  {
    int nValue  = 0;

    // Make sure camera is in reset mode
    if ( !Reset() )
    {
      return false;
    }

    // Calculate configuration value for given fRGB and fAutoWhite
    if ( fRGB )
    {
      nValue = 40;
    }
    else
    {
      nValue = 32;
    }
    if ( fAutoWhite )
    {
      nValue += 4;
    }

    // Configure camera
    m_buffTemp.clear();
    m_buffTemp.append("CR 18 ");
    m_buffTemp.append(nValue);
    m_buffTemp.append("\r");

    m_uartTx.sendString(m_buffTemp.toString());
    if ( !WaitForACKPrompt() )
    {
      return false;
    }
    m_fRGB        = fRGB;
    m_fAutoWhite  = fAutoWhite;

    return true;
  }

  public boolean SetFrameSpeed(int nFrameSpeed)
  {
    // Make sure camera is in reset mode
    if ( !Reset() )
    {
      return false;
    }

    // Configure camera contrast
    m_buffTemp.clear();
    m_buffTemp.append("CR 17 ");
    m_buffTemp.append(nFrameSpeed);
    m_buffTemp.append("\r");

    m_uartTx.sendString(m_buffTemp.toString());
    if ( !WaitForACKPrompt() )
    {
      return false;
    }

    m_nFrameSpeed = nFrameSpeed;
    return true;
  }

  public boolean SetAutoGain(boolean fAutoGain)
  {
    int nValue  = 0;

    // Make sure camera is in reset mode
    if ( !Reset() )
    {
      return false;
    }

    // Calculate configuration value for given fRGB and fAutoWhite
    if ( fAutoGain )
    {
      nValue = 33;
    }
    else
    {
      nValue = 32;
    }

    // Configure camera
    m_buffTemp.clear();
    m_buffTemp.append("CR 19 ");
    m_buffTemp.append(nValue);
    m_buffTemp.append("\r");

    m_uartTx.sendString(m_buffTemp.toString());
    if ( !WaitForACKPrompt() )
    {
      return false;
    }

    return true;
  }

  // Manually configure register of the OV6620 camera
  // This method shouldn't be used for any of the standard
  // settings for which other methods exist since this object
  // tracks settings like frame rate, colour mode, etc.
  public boolean ConfigRegister(int nRegister, int nValue)
  {
    // Make sure camera is in reset mode
    if ( !Reset() )
    {
      return false;
    }

    // Configure camera
    m_buffTemp.clear();
    m_buffTemp.append("CR ");
    m_buffTemp.append(nRegister);
    m_buffTemp.append(" ");
    m_buffTemp.append(nValue);
    m_buffTemp.append("\r");

    m_uartTx.sendString(m_buffTemp.toString());
    if ( !WaitForACKPrompt() )
    {
      return false;
    }

    // Check for special register values that are tracked by this
    // class
    if ( nRegister == 18 )
    {
      // User can set Auto White Balance and RGB mode with register 18
      if ( (nValue & 4 ) == 4 )
      {
        // Auto White Balance is turned on
        m_fAutoWhite = true;
      }
      else
      {
        m_fAutoWhite = false;
      }
      if ( (nValue & 8 ) == 8 )
      {
        m_fRGB = true;
      }
      else
      {
        m_fRGB = false;
      }
    }
    if ( nRegister == 17 )
    {
      // User can set frame speed with register 17
      m_nFrameSpeed = nValue;
    }

    return true;
  }

  // Dumps 80 x 36 B&W Frame
  // There will be a single intensity byte stored for each pixel.
  // The pixels are stored in column first order.  This means that
  // the rightmost column of pixels is stored in the first 36 bytes
  // of the array.  The next 36 bytes are the pixels for the next
  // column to the left.
  public boolean DumpBWFrame(char rgchBuff[])
  {
    int     iRow          = 0;
    int     iCol          = 0;
    int     iVal          = 0;
    int     iByte         = 0;
    char    chByte        = 0;
    boolean fRGB          = false;
    boolean fAutoWhite    = false;
    int     nFrameSpeed   = 0;

    // Make sure the caller's buffer is long enough
    if ( rgchBuff.length < 80 * 36 )
    {
      return false;
    }

    // Make sure camera is in idle mode
    if ( !Reset() )
    {
      return false;
    }

    // Remember the camera's state
    fRGB        = m_fRGB;
    fAutoWhite  = m_fAutoWhite;
    nFrameSpeed = m_nFrameSpeed;

    //UNDONE: Should probably reset WindowSize as well
    try
    {
      // Set the camera to CrYCb
      if ( !SetColorMode(false, false) )
      {
        return false;
      }

      // Slow frame rate down to minimum
      if ( !SetFrameSpeed(FRAME_SPEED_4) )
      {
        return false;
      }

      // Add a delay to each packet to slow it down
      m_uartTx.sendString("DM 22\r");
      if ( !WaitForACKPrompt() )
      {
        return false;
      }

      // Tell camera to start dumping frame
      m_uartTx.sendString("DF\r");
      if ( !WaitForACK() )
      {
        return false;
      }

      // Read the data from the camera and store it in the caller
      // supplied buffer
      while ( true )
      {
        chByte = (char) m_uartRx.receiveByte();
        if ( chByte == 0x02 )
        {
          // New Column
          if ( iRow != 36 )
          {
            break;
          }
          iRow  = 0;
          iVal  = 0;
          iCol++;
        }
        else if ( chByte == 0x03 )
        {
          // End of frame
          break;
        }
        else if ( chByte != 0x02 )
        {
          if ( iVal % 3 == 1 )
          {
            // This is the Y (intensity) component for this row
            if ( iByte >= rgchBuff.length )
            {
              break;
            }
            rgchBuff[iByte++] = chByte;
            iRow++;
          }
          iVal++;
        }
      }
      WaitForPrompt();
      if ( iByte != 80 * 36 )
      {
        return false;
      }

      return true;
    }
    finally
    {
          // UNDONE: Remove this debug code
          System.out.print("End of frame (");
          System.out.print(iCol);
          System.out.print(",");
          System.out.print(iByte);
          System.out.print(",");
          System.out.print(iRow);
          System.out.println(")");

      // Set camera back to original camera mode
      m_uartTx.sendString("DM 0\r");
      WaitForACKPrompt();
      SetFrameSpeed(m_nFrameSpeed);
      SetColorMode(m_fRGB, m_fAutoWhite);
    }
  }

  // Allows user to set the window to be used by commands such as
  // tracking.  The valid X values are 1 - 80 and the valid Y
  // values are 1 - 143
  boolean SetWindow(int nX1, int nY1, int nX2, int nY2)
  {
    // Make sure camera is in reset mode
    if ( !Reset() )
    {
      return false;
    }

    // Configure camera window size
    m_buffTemp.clear();
    m_buffTemp.append("SW ");
    m_buffTemp.append(nX1);
    m_buffTemp.append(" ");
    m_buffTemp.append(nY1);
    m_buffTemp.append(" ");
    m_buffTemp.append(nX2);
    m_buffTemp.append(" ");
    m_buffTemp.append(nY2);
    m_buffTemp.append("\r");

    m_uartTx.sendString(m_buffTemp.toString());
    if ( !WaitForACKPrompt() )
    {
      return false;
    }

    // Save the window state
    m_nX1 = nX1;
    m_nY1 = nY1;
    m_nX2 = nX2;
    m_nY2 = nY2;

    return true;
  }

  // Get mean and deviation of the colors within the current window
  boolean GetStats(CMUCamStatData stats)
  {
    int             nVal    = 0;

    if ( m_nTrackingMode != TRACKING_STATS )
    {
      // Make sure camera is in reset mode
      if ( !Reset() )
      {
        return false;
      }

      // Tell camera to start dumping statistics
      m_uartTx.sendString("GM\r");
      if ( !WaitForACK() )
      {
        return false;
      }

      m_nTrackingMode = TRACKING_STATS;
    }
    else
    {
      // Clear out the buffer on the RX Uart
      m_uartRx.stop();
      m_uartRx.start();
    }

    // Wait for the 0xFF start of packet byte
    while ( 0xFF != (0xFF & m_uartRx.receiveByte()) );

    // Validate that this is a S packet
    nVal = 0xFF & m_uartRx.receiveByte();
    if ( nVal != 'S' )
    {
      return false;
    }

    // Retrieve mean and deviation values
    stats.m_nRmean = 0xFF & m_uartRx.receiveByte();
    stats.m_nGmean = 0xFF & m_uartRx.receiveByte();
    stats.m_nBmean = 0xFF & m_uartRx.receiveByte();
    stats.m_nRdev  = 0xFF & m_uartRx.receiveByte();
    stats.m_nGdev  = 0xFF & m_uartRx.receiveByte();
    stats.m_nBdev  = 0xFF & m_uartRx.receiveByte();
    stats.m_fValid = true;

    return true;
  }

  // Track Coloured Object
  boolean TrackColor (int nRmin, int nRmax,
                      int nGmin, int nGmax,
                      int nBmin, int nBmax,
                      CMUCamTrackingData tracking)
  {
    int            nVal    = 0;

    if ( m_nTrackingMode != TRACKING_COLOR )
    {
      // Make sure camera is in reset mode
      if ( !Reset() )
      {
        return false;
      }

      // Disable Middle Mass mode
      m_uartTx.sendString("MM 0\r");
      if ( !WaitForACKPrompt() )
      {
        return false;
      }

      // Tell camera to start tracking coloured object
      m_buffTemp.clear();
      m_buffTemp.append("TC ");
      m_buffTemp.append(nRmin);
      m_buffTemp.append(" ");
      m_buffTemp.append(nRmax);
      m_buffTemp.append(" ");
      m_buffTemp.append(nGmin);
      m_buffTemp.append(" ");
      m_buffTemp.append(nGmax);
      m_buffTemp.append(" ");
      m_buffTemp.append(nBmin);
      m_buffTemp.append(" ");
      m_buffTemp.append(nBmax);
      m_buffTemp.append("\r");
      m_uartTx.sendString(m_buffTemp.toString());
      if ( !WaitForACK() )
      {
        return false;
      }

      m_nTrackingMode = TRACKING_COLOR;
    }
    else
    {
      // Clear out the buffer on the RX Uart
      m_uartRx.stop();
      m_uartRx.start();
    }

    // Wait for the 0xFF start of packet byte
    while ( 0xFF != (0xFF & m_uartRx.receiveByte()) );

    // Validate that this is a C packet
    nVal = 0xFF & m_uartRx.receiveByte();
    if ( nVal != 'C' )
    {
      return false;
    }

    // Retrieve coordinates
    tracking.m_nX1         = 0xFF & m_uartRx.receiveByte();
    tracking.m_nY1         = 0xFF & m_uartRx.receiveByte();
    tracking.m_nX2         = 0xFF & m_uartRx.receiveByte();
    tracking.m_nY2         = 0xFF & m_uartRx.receiveByte();
    tracking.m_nPixels     = 0xFF & m_uartRx.receiveByte();
    tracking.m_nConfidence = 0xFF & m_uartRx.receiveByte();
    tracking.m_fValid      = true;

    return true;
  }

  // Track Coloured Object w/ Middle of Mass
  boolean TrackColor (int nRmin, int nRmax,
                      int nGmin, int nGmax,
                      int nBmin, int nBmax,
                      CMUCamTrackingData tracking,
                      CMUCamMassData mass)
  {
    int            nVal    = 0;

    if ( m_nTrackingMode != TRACKING_COLOR_MASS )
    {
      // Make sure camera is in reset mode
      if ( !Reset() )
      {
        return false;
      }

      // Enable Middle Mass mode
      m_uartTx.sendString("MM 1\r");
      if ( !WaitForACKPrompt() )
      {
        return false;
      }

      // Tell camera to start tracking coloured object
      m_buffTemp.clear();
      m_buffTemp.append("TC ");
      m_buffTemp.append(nRmin);
      m_buffTemp.append(" ");
      m_buffTemp.append(nRmax);
      m_buffTemp.append(" ");
      m_buffTemp.append(nGmin);
      m_buffTemp.append(" ");
      m_buffTemp.append(nGmax);
      m_buffTemp.append(" ");
      m_buffTemp.append(nBmin);
      m_buffTemp.append(" ");
      m_buffTemp.append(nBmax);
      m_buffTemp.append("\r");
      m_uartTx.sendString(m_buffTemp.toString());
      if ( !WaitForACK() )
      {
        return false;
      }

      m_nTrackingMode = TRACKING_COLOR_MASS;
    }
    else
    {
      // Clear out the buffer on the RX Uart
      m_uartRx.stop();
      m_uartRx.start();
    }

    // Wait for the 0xFF start of packet byte
    while ( 0xFF != (0xFF & m_uartRx.receiveByte()) );

    // Validate that this is a M packet
    nVal = 0xFF & m_uartRx.receiveByte();
    if ( nVal != 'M' )
    {
      return false;
    }

    // Retrieve coordinates
    mass.m_nX              = 0xFF & m_uartRx.receiveByte();
    mass.m_nY              = 0xFF & m_uartRx.receiveByte();
    tracking.m_nX1         = 0xFF & m_uartRx.receiveByte();
    tracking.m_nY1         = 0xFF & m_uartRx.receiveByte();
    tracking.m_nX2         = 0xFF & m_uartRx.receiveByte();
    tracking.m_nY2         = 0xFF & m_uartRx.receiveByte();
    tracking.m_nPixels     = 0xFF & m_uartRx.receiveByte();
    tracking.m_nConfidence = 0xFF & m_uartRx.receiveByte();
    tracking.m_fValid      = true;

    return true;
  }

  // TrackObject w/ Bitmap
  // TrackObject w/ Stats

  // Get/Set Servo Pin Data/
  // Set Tracking Light/
  // Set Noise Filter
  // GetVersion
/*
    // Display camera version to debug terminal
    m_uartTx.sendString("GV\r");
    if ( !WaitForACK() )
    {
      return false;
    }
    for ( ch = ' ' ; ch != ':' ; ch = (char)m_uartRx.receiveByte() )
    {
      System.out.print(ch);
    }
    System.out.println("");
    m_uartRx.receiveByte();
*/


  // Wait for camera to return prompt
  protected boolean WaitForPrompt()
  {
    char    ch = 0;

    // Retrieve last character sent back from camera
    m_timer.mark();
    while ( !m_timer.timeout(ACK_TIMEOUT)  )
    {
      // If there is a byte available, get it and store it
      if ( m_uartRx.byteAvailable() )
      {
        ch = (char) (0xff & m_uartRx.receiveByte());
        m_timer.mark();
      }
    }

    // Was the ":" prompt encountered
    if (  ch == ':'  )
    {
      return true;
    }
    else
    {
      return false;
    }
  }

  // Wait for camera to finish with a ACK packet and Prompt
  protected boolean WaitForACKPrompt()
  {
    ClearBuffer();

    // Retrieve last 5 characters sent back from camera
    m_timer.mark();
    while ( !m_timer.timeout(ACK_TIMEOUT)  )
    {
      // If there is a byte available, get it and store it along
      // with the last 4 bytes retrieved.
      if ( m_uartRx.byteAvailable() )
      {
        m_rgchBuff[0] = m_rgchBuff[1];
        m_rgchBuff[1] = m_rgchBuff[2];
        m_rgchBuff[2] = m_rgchBuff[3];
        m_rgchBuff[3] = m_rgchBuff[4];
        m_rgchBuff[4] = (char) (0xff & m_uartRx.receiveByte());
        m_timer.mark();
      }
    }

    // Were the last 5 characters "ACK\r:"
    if (  m_rgchBuff[0] == 'A' &&
          m_rgchBuff[1] == 'C' &&
          m_rgchBuff[2] == 'K' &&
          m_rgchBuff[3] == '\r'&&
          m_rgchBuff[4] == ':'  )
    {
      return true;
    }
    else
    {
      return false;
    }
  }

  // Wait for camera to return ACK packet
  protected boolean WaitForACK()
  {
    int     i       = 0;

    ClearBuffer();

    // Retrieve next 4 characters sent back from camera
    m_timer.mark();
    while ( i < 4 && !m_timer.timeout(ACK_TIMEOUT) )
    {
      // If there is a byte available, get it and store it along
      // with the last 3 bytes retrieved.
      if ( m_uartRx.byteAvailable() )
      {
        m_rgchBuff[0] = m_rgchBuff[1];
        m_rgchBuff[1] = m_rgchBuff[2];
        m_rgchBuff[2] = m_rgchBuff[3];
        m_rgchBuff[3] = (char) m_uartRx.receiveByte();
        m_timer.mark();
        i++;
      }
    }

    // Were the last 4 characters "ACK\r"
    if (  m_rgchBuff[0] == 'A' &&
          m_rgchBuff[1] == 'C' &&
          m_rgchBuff[2] == 'K' &&
          m_rgchBuff[3] == '\r' )
    {
      return true;
    }
    else
    {
      return false;
    }
  }

  protected void ClearBuffer()
  {
    int i = 0;

    for ( i = 0 ; i < m_rgchBuff.length ; i++ )
    {
      m_rgchBuff[i] = 0;
    }
  }

  // Uart objects used for communicating to CMUCam.  One for receiving
  // and the other for transmitting.
  Uart              m_uartRx;
  Uart              m_uartTx;

  // Timer used to timeout waiting for bytes from camera
  Timer             m_timer;

  // Temporary string buffer used for converting int, etc to string
  StringBuffer      m_buffTemp;
  char              m_rgchBuff[];

  // State member variables
  boolean           m_fInit;
  boolean           m_fRGB;
  boolean           m_fAutoWhite;
  int               m_nFrameSpeed;
  int               m_nX1;
  int               m_nY1;
  int               m_nX2;
  int               m_nY2;
  int               m_nTrackingMode;

  // This class fixes the baud rate at 19200.  There are two higher
  // speeds supported by the Javelin but they don't have enough
  // oversampling to reliably read data (ie. bits get flipped at higher
  // speeds.)
  static final int  BAUD_RATE     = Uart.speed19200;
  static final int  STOP_BITS     = Uart.stop1;
  static final int  ACK_TIMEOUT   = 100;  // wait X milliseconds for next byte

  // Frame speeds to be passed into SetFrameSpeed
  public static final int FRAME_SPEED_17  = 2;
  public static final int FRAME_SPEED_13  = 3;
  public static final int FRAME_SPEED_11  = 4;
  public static final int FRAME_SPEED_9   = 5;
  public static final int FRAME_SPEED_8   = 6;
  public static final int FRAME_SPEED_7   = 7;
  public static final int FRAME_SPEED_6   = 8;
  public static final int FRAME_SPEED_5   = 10;
  public static final int FRAME_SPEED_4   = 12;

  // Tracking modes
  static final int TRACKING_NONE       = 0;
  static final int TRACKING_STATS      = 1;
  static final int TRACKING_COLOR      = 2;
  static final int TRACKING_COLOR_MASS = 3;
}